# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import asyncio.subprocess
import concurrent.futures as futures
import contextlib
import itertools
import logging
import multiprocessing
import os
import threading
import traceback
from abc import ABC, ABCMeta, abstractmethod
from typing import Any, Callable, Coroutine, Optional, Type, TypeVar

import psutil

from .._utils import TypeDispatcher, create_actor_ref, to_binary
from ..api import Actor
from ..core import ActorRef, BufferRef, FileObjectRef, register_local_pool
from ..debug import debug_async_timeout, record_message_trace
from ..errors import (
    ActorAlreadyExist,
    ActorNotExist,
    CannotCancelTask,
    SendMessageFailed,
    ServerClosed,
)
from ..metrics import init_metrics
from ..utils import implements, is_zero_ip, register_asyncio_task_timeout_detector
from .allocate_strategy import AddressSpecified, allocated_type
from .communication import (
    Channel,
    Server,
    UCXChannel,
    gen_local_address,
    get_server_type,
)
from .communication.errors import ChannelClosed
from .config import ActorPoolConfig
from .core import ActorCaller, ResultMessageType
from .message import (
    DEFAULT_PROTOCOL,
    ActorRefMessage,
    CancelMessage,
    ControlMessage,
    ControlMessageType,
    CreateActorMessage,
    DestroyActorMessage,
    ErrorMessage,
    ForwardMessage,
    HasActorMessage,
    MessageType,
    ResultMessage,
    SendMessage,
    TellMessage,
    _MessageBase,
    new_message_id,
)
from .router import Router

logger = logging.getLogger(__name__)


@contextlib.contextmanager
def _disable_log_temporally():
    if os.getenv("CUDA_VISIBLE_DEVICES") == "-1":
        # disable logging when CUDA_VISIBLE_DEVICES == -1
        # many logging comes from ptxcompiler may distract users
        try:
            logging.disable(level=logging.ERROR)
            yield
        finally:
            logging.disable(level=logging.NOTSET)
    else:
        yield


class _ErrorProcessor:
    def __init__(self, address: str, message_id: bytes, protocol):
        self._address = address
        self._message_id = message_id
        self._protocol = protocol
        self.result = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.result is None:
            self.result = ErrorMessage(
                self._message_id,
                self._address,
                os.getpid(),
                exc_type,
                exc_val,
                exc_tb,
                protocol=self._protocol,
            )
            return True


def _register_message_handler(pool_type: Type["AbstractActorPool"]):
    pool_type._message_handler = dict()
    for message_type, handler in [
        (MessageType.create_actor, pool_type.create_actor),
        (MessageType.destroy_actor, pool_type.destroy_actor),
        (MessageType.has_actor, pool_type.has_actor),
        (MessageType.actor_ref, pool_type.actor_ref),
        (MessageType.send, pool_type.send),
        (MessageType.tell, pool_type.tell),
        (MessageType.cancel, pool_type.cancel),
        (MessageType.forward, pool_type.forward),
        (MessageType.control, pool_type.handle_control_command),
        (MessageType.copy_to_buffers, pool_type.handle_copy_to_buffers_message),
        (MessageType.copy_to_fileobjs, pool_type.handle_copy_to_fileobjs_message),
    ]:
        pool_type._message_handler[message_type] = handler  # type: ignore
    return pool_type


class AbstractActorPool(ABC):
    __slots__ = (
        "process_index",
        "label",
        "external_address",
        "internal_address",
        "env",
        "_servers",
        "_router",
        "_config",
        "_stopped",
        "_actors",
        "_caller",
        "_process_messages",
        "_asyncio_task_timeout_detector_task",
    )

    _message_handler: dict[MessageType, Callable]
    _process_messages: dict[bytes, asyncio.Future | asyncio.Task | None]

    def __init__(
        self,
        process_index: int,
        label: str,
        external_address: str,
        internal_address: str,
        env: dict,
        router: Router,
        config: ActorPoolConfig,
        servers: list[Server],
    ):
        # register local pool for local actor lookup.
        # The pool is weakrefed, so we don't need to unregister it.
        if not is_zero_ip(external_address):
            # Only register_local_pool when we listen on non-zero ip (because all-zero ip is wildcard address),
            # avoid mistaken with another remote service listen on non-zero ip with the same port.
            register_local_pool(external_address, self)
        self.process_index = process_index
        self.label = label
        self.external_address = external_address
        self.internal_address = internal_address
        self.env = env
        self._router = router
        self._config = config
        self._servers = servers

        self._stopped = asyncio.Event()

        # states
        # actor id -> actor
        self._actors: dict[bytes, Actor] = dict()
        # message id -> future
        self._process_messages = dict()

        # manage async actor callers
        self._caller = ActorCaller()
        self._asyncio_task_timeout_detector_task = (
            register_asyncio_task_timeout_detector()
        )
        # init metrics
        metric_configs = self._config.get_metric_configs()
        metric_backend = metric_configs.get("backend")
        init_metrics(metric_backend, config=metric_configs.get(metric_backend))

    @property
    def router(self):
        return self._router

    @abstractmethod
    async def create_actor(self, message: CreateActorMessage) -> ResultMessageType:
        """
        Create an actor.

        Parameters
        ----------
        message: CreateActorMessage
            message to create an actor.

        Returns
        -------
        result_message
            result or error message.
        """

    @abstractmethod
    async def has_actor(self, message: HasActorMessage) -> ResultMessage:
        """
        Check if an actor exists or not.

        Parameters
        ----------
        message: HasActorMessage
            message

        Returns
        -------
        result_message
            result message contains if an actor exists or not.
        """

    @abstractmethod
    async def destroy_actor(self, message: DestroyActorMessage) -> ResultMessageType:
        """
        Destroy an actor.

        Parameters
        ----------
        message: DestroyActorMessage
            message to destroy an actor.

        Returns
        -------
        result_message
            result or error message.
        """

    @abstractmethod
    async def actor_ref(self, message: ActorRefMessage) -> ResultMessageType:
        """
        Get an actor's ref.

        Parameters
        ----------
        message: ActorRefMessage
            message to get an actor's ref.

        Returns
        -------
        result_message
            result or error message.
        """

    @abstractmethod
    async def send(self, message: SendMessage) -> ResultMessageType:
        """
        Send a message to some actor.

        Parameters
        ----------
        message: SendMessage
            Message to send.

        Returns
        -------
        result_message
            result or error message.
        """

    @abstractmethod
    async def tell(self, message: TellMessage) -> ResultMessageType:
        """
        Tell message to some actor.

        Parameters
        ----------
        message: TellMessage
            Message to tell.

        Returns
        -------
        result_message
            result or error message.
        """

    @abstractmethod
    async def cancel(self, message: CancelMessage) -> ResultMessageType:
        """
        Cancel message that sent

        Parameters
        ----------
        message: CancelMessage
            Cancel message.

        Returns
        -------
        result_message
            result or error message
        """

    async def forward(self, message: ForwardMessage) -> ResultMessageType:
        """
        Forward message

        Parameters
        ----------
        message: ForwardMessage
            Forward message.

        Returns
        -------
        result_message
            result or error message
        """
        return await self.call(message.address, message.raw_message)

    def _sync_pool_config(self, actor_pool_config: ActorPoolConfig):
        self._config = actor_pool_config
        # remove router from global one
        global_router = Router.get_instance()
        global_router.remove_router(self._router)  # type: ignore
        # update router
        self._router.set_mapping(actor_pool_config.external_to_internal_address_map)
        # update global router
        global_router.add_router(self._router)  # type: ignore

    async def handle_control_command(
        self, message: ControlMessage
    ) -> ResultMessageType:
        """
        Handle control command.

        Parameters
        ----------
        message: ControlMessage
            Control message.

        Returns
        -------
        result_message
            result or error message.
        """
        with _ErrorProcessor(
            self.external_address, message.message_id, protocol=message.protocol
        ) as processor:
            content: bool | ActorPoolConfig = True
            if message.control_message_type == ControlMessageType.stop:
                await self.stop()
            elif message.control_message_type == ControlMessageType.sync_config:
                self._sync_pool_config(message.content)
            elif message.control_message_type == ControlMessageType.get_config:
                if message.content == "main_pool_address":
                    main_process_index = self._config.get_process_indexes()[0]
                    content = self._config.get_pool_config(main_process_index)[
                        "external_address"
                    ][0]
                else:
                    content = self._config
            else:  # pragma: no cover
                raise TypeError(
                    f"Unable to handle control message "
                    f"with type {message.control_message_type}"
                )
            processor.result = ResultMessage(
                message.message_id, content, protocol=message.protocol
            )

        return processor.result

    async def _run_coro(self, message_id: bytes, coro: Coroutine):
        self._process_messages[message_id] = asyncio.tasks.current_task()
        try:
            return await coro
        finally:
            self._process_messages.pop(message_id, None)

    async def _send_channel(
        self, result: _MessageBase, channel: Channel, resend_failure: bool = True
    ):
        try:
            await channel.send(result)
        except (ChannelClosed, ConnectionResetError):
            if not self._stopped.is_set() and not channel.closed:
                raise
        except Exception as ex:
            logger.exception(
                "Error when sending message %s from %s to %s",
                result.message_id.hex(),
                channel.local_address,
                channel.dest_address,
            )
            if not resend_failure:  # pragma: no cover
                raise

            with _ErrorProcessor(
                self.external_address, result.message_id, result.protocol
            ) as processor:
                error_msg = (
                    f"Error when sending message {result.message_id.hex()}. "
                    f"Caused by {ex!r}. "
                )
                if isinstance(result, ErrorMessage):
                    format_tb = "\n".join(traceback.format_tb(result.traceback))
                    error_msg += (
                        f"\nOriginal error: {result.error!r}"
                        f"Traceback: \n{format_tb}"
                    )
                else:
                    error_msg += "See server logs for more details"
                raise SendMessageFailed(error_msg) from None
            await self._send_channel(processor.result, channel, resend_failure=False)

    async def process_message(self, message: _MessageBase, channel: Channel):
        handler = self._message_handler[message.message_type]
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            # use `%.500` to avoid print too long messages
            with debug_async_timeout(
                "process_message_timeout",
                "Process message %.500s of channel %s timeout.",
                message,
                channel,
            ):
                processor.result = await self._run_coro(
                    message.message_id, handler(self, message)
                )

        await self._send_channel(processor.result, channel)

    async def call(self, dest_address: str, message: _MessageBase) -> ResultMessageType:
        return await self._caller.call(self._router, dest_address, message)  # type: ignore

    @staticmethod
    def _parse_config(config: dict, kw: dict) -> dict:
        actor_pool_config: ActorPoolConfig = config.pop("actor_pool_config")
        kw["config"] = actor_pool_config
        kw["process_index"] = process_index = config.pop("process_index")
        curr_pool_config = actor_pool_config.get_pool_config(process_index)
        kw["label"] = curr_pool_config["label"]
        external_addresses = curr_pool_config["external_address"]
        kw["external_address"] = external_addresses[0]
        kw["internal_address"] = curr_pool_config["internal_address"]
        kw["router"] = Router(
            external_addresses,
            gen_local_address(process_index),
            actor_pool_config.external_to_internal_address_map,
            comm_config=actor_pool_config.get_comm_config(),
            proxy_config=actor_pool_config.get_proxy_config(),
        )
        kw["env"] = curr_pool_config["env"]

        if config:  # pragma: no cover
            raise TypeError(
                f"Creating pool got unexpected " f'arguments: {",".join(config)}'
            )

        return kw

    @classmethod
    @abstractmethod
    async def create(cls, config: dict) -> "AbstractActorPool":
        """
        Create an actor pool.

        Parameters
        ----------
        config: dict
            configurations.

        Returns
        -------
        actor_pool:
            Actor pool.
        """

    async def start(self):
        if self._stopped.is_set():
            raise RuntimeError("pool has been stopped, cannot start again")
        start_servers = [server.start() for server in self._servers]
        await asyncio.gather(*start_servers)

    async def join(self, timeout: float | None = None):
        wait_stopped = asyncio.create_task(self._stopped.wait())

        try:
            await asyncio.wait_for(wait_stopped, timeout=timeout)
        except (futures.TimeoutError, asyncio.TimeoutError):  # pragma: no cover
            wait_stopped.cancel()

    async def stop(self):
        try:
            # clean global router
            router = Router.get_instance()
            if router is not None:
                router.remove_router(self._router)
            stop_tasks = []
            # stop all servers
            stop_tasks.extend([server.stop() for server in self._servers])
            # stop all clients
            stop_tasks.append(self._caller.stop())
            await asyncio.gather(*stop_tasks)

            self._servers = []
            if self._asyncio_task_timeout_detector_task:  # pragma: no cover
                self._asyncio_task_timeout_detector_task.cancel()
        finally:
            self._stopped.set()

    async def handle_copy_to_buffers_message(self, message) -> ResultMessage:
        for addr, uid, start, _len, data in message.content:
            buffer = BufferRef.get_buffer(BufferRef(addr, uid))
            buffer[start : start + _len] = data
        return ResultMessage(message_id=message.message_id, result=True)

    async def handle_copy_to_fileobjs_message(self, message) -> ResultMessage:
        for addr, uid, data in message.content:
            file_obj = FileObjectRef.get_local_file_object(FileObjectRef(addr, uid))
            await file_obj.write(data)
        return ResultMessage(message_id=message.message_id, result=True)

    @property
    def stopped(self) -> bool:
        return self._stopped.is_set()

    async def _handle_ucx_meta_message(
        self, message: _MessageBase, channel: Channel
    ) -> bool:
        if (
            isinstance(message, ControlMessage)
            and message.message_type == MessageType.control
            and message.control_message_type == ControlMessageType.switch_to_copy_to
            and isinstance(channel, UCXChannel)
        ):
            with _ErrorProcessor(
                self.external_address, message.message_id, message.protocol
            ) as processor:
                # use `%.500` to avoid print too long messages
                with debug_async_timeout(
                    "process_message_timeout",
                    "Process message %.500s of channel %s timeout.",
                    message,
                    channel,
                ):
                    buffers = [
                        BufferRef.get_buffer(BufferRef(addr, uid))
                        for addr, uid in message.content
                    ]
                    await channel.recv_buffers(buffers)
                    processor.result = ResultMessage(
                        message_id=message.message_id, result=True
                    )
            asyncio.create_task(self._send_channel(processor.result, channel))
            return True
        return False

    async def on_new_channel(self, channel: Channel):
        try:
            while not self._stopped.is_set():
                try:
                    message = await channel.recv()
                except (EOFError, ConnectionError, BrokenPipeError) as e:
                    logger.debug(f"pool: close connection due to {e}")
                    # no data to read, check channel
                    try:
                        await channel.close()
                    except (ConnectionError, EOFError):
                        # close failed, ignore
                        pass
                    return
                if await self._handle_ucx_meta_message(message, channel):
                    continue
                asyncio.create_task(self.process_message(message, channel))
                # delete to release the reference of message
                del message
                await asyncio.sleep(0)
        finally:
            try:
                await channel.close()
            except:  # noqa: E722  # nosec  # pylint: disable=bare-except
                # ignore all error if fail to close at last
                pass

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()


class ActorPoolBase(AbstractActorPool, metaclass=ABCMeta):
    __slots__ = ()

    @implements(AbstractActorPool.create_actor)
    async def create_actor(self, message: CreateActorMessage) -> ResultMessageType:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            actor_id = message.actor_id
            if actor_id in self._actors:
                raise ActorAlreadyExist(
                    f"Actor {actor_id!r} already exist, cannot create"
                )

            actor = message.actor_cls(*message.args, **message.kwargs)
            actor.uid = actor_id
            actor.address = address = self.external_address
            self._actors[actor_id] = actor
            await self._run_coro(message.message_id, actor.__post_create__())

            proxies = self._router.get_proxies(address)
            result = ActorRef(address, actor_id, proxy_addresses=proxies)
            # ensemble result message
            processor.result = ResultMessage(
                message.message_id, result, protocol=message.protocol
            )
        return processor.result

    @implements(AbstractActorPool.has_actor)
    async def has_actor(self, message: HasActorMessage) -> ResultMessage:
        result = ResultMessage(
            message.message_id,
            message.actor_ref.uid in self._actors,
            protocol=message.protocol,
        )
        return result

    @implements(AbstractActorPool.destroy_actor)
    async def destroy_actor(self, message: DestroyActorMessage) -> ResultMessageType:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            actor_id = message.actor_ref.uid
            try:
                actor = self._actors[actor_id]
            except KeyError:
                raise ActorNotExist(f"Actor {actor_id} does not exist")
            await self._run_coro(message.message_id, actor.__pre_destroy__())
            del self._actors[actor_id]

            processor.result = ResultMessage(
                message.message_id, actor_id, protocol=message.protocol
            )
        return processor.result

    @implements(AbstractActorPool.actor_ref)
    async def actor_ref(self, message: ActorRefMessage) -> ResultMessageType:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            actor_id = message.actor_ref.uid
            if actor_id not in self._actors:
                raise ActorNotExist(f"Actor {actor_id} does not exist")
            proxies = self._router.get_proxies(self.external_address)
            result = ResultMessage(
                message.message_id,
                ActorRef(self.external_address, actor_id, proxy_addresses=proxies),
                protocol=message.protocol,
            )
            processor.result = result
        return processor.result

    @implements(AbstractActorPool.send)
    async def send(self, message: SendMessage) -> ResultMessageType:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor, record_message_trace(message):
            actor_id = message.actor_ref.uid
            if actor_id not in self._actors:
                raise ActorNotExist(f"Actor {actor_id} does not exist")
            coro = self._actors[actor_id].__on_receive__(message.content)
            result = await self._run_coro(message.message_id, coro)
            processor.result = ResultMessage(
                message.message_id,
                result,
                protocol=message.protocol,
                profiling_context=message.profiling_context,
            )
        return processor.result

    @implements(AbstractActorPool.tell)
    async def tell(self, message: TellMessage) -> ResultMessageType:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            actor_id = message.actor_ref.uid
            if actor_id not in self._actors:  # pragma: no cover
                raise ActorNotExist(f"Actor {actor_id} does not exist")
            call = self._actors[actor_id].__on_receive__(message.content)
            # asynchronously run, tell does not care about result
            asyncio.create_task(call)
            await asyncio.sleep(0)
            processor.result = ResultMessage(
                message.message_id,
                None,
                protocol=message.protocol,
                profiling_context=message.profiling_context,
            )
        return processor.result

    @implements(AbstractActorPool.cancel)
    async def cancel(self, message: CancelMessage) -> ResultMessageType:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            future = self._process_messages.get(message.cancel_message_id)
            if future is None or future.done():  # pragma: no cover
                raise CannotCancelTask(
                    "Task not exists, maybe it is done or cancelled already"
                )
            future.cancel()
            processor.result = ResultMessage(
                message.message_id, True, protocol=message.protocol
            )
        return processor.result

    @staticmethod
    def _set_global_router(router: Router):
        # be cautious about setting global router
        # for instance, multiple main pool may be created in the same process

        # get default router or create an empty one
        default_router = Router.get_instance_or_empty()
        Router.set_instance(default_router)
        # append this router to global
        default_router.add_router(router)

    @staticmethod
    def _update_stored_addresses(
        servers: list[Server],
        raw_addresses: list[str],
        actor_pool_config: ActorPoolConfig,
        kw: dict,
    ):
        process_index = kw["process_index"]
        curr_pool_config = actor_pool_config.get_pool_config(process_index)
        external_addresses = curr_pool_config["external_address"]
        external_address_set = set(external_addresses)

        kw["servers"] = servers

        new_external_addresses = [
            server.address
            for server, raw_address in zip(servers, raw_addresses)
            if raw_address in external_address_set
        ]

        if external_address_set != set(new_external_addresses):
            external_addresses = new_external_addresses
            actor_pool_config.reset_pool_external_address(
                process_index, external_addresses
            )
            external_addresses = curr_pool_config["external_address"]

            logger.debug(
                "External address of process index %s updated to %s",
                process_index,
                external_addresses[0],
            )
            if kw["internal_address"] == kw["external_address"]:
                # internal address may be the same as external address in Windows
                kw["internal_address"] = external_addresses[0]
            kw["external_address"] = external_addresses[0]

            kw["router"] = Router(
                external_addresses,
                gen_local_address(process_index),
                actor_pool_config.external_to_internal_address_map,
                comm_config=actor_pool_config.get_comm_config(),
                proxy_config=actor_pool_config.get_proxy_config(),
            )

    @classmethod
    async def _create_servers(
        cls, addresses: list[str], channel_handler: Callable, config: dict
    ):
        assert len(set(addresses)) == len(addresses)
        # create servers
        create_server_tasks = []
        for addr in addresses:
            server_type = get_server_type(addr)
            extra_config = server_type.parse_config(config)
            server_config = dict(address=addr, handle_channel=channel_handler)
            server_config.update(extra_config)
            task = asyncio.create_task(server_type.create(server_config))
            create_server_tasks.append(task)

        await asyncio.gather(*create_server_tasks)
        return [f.result() for f in create_server_tasks]

    @classmethod
    @implements(AbstractActorPool.create)
    async def create(cls, config: dict) -> "ActorPoolType":
        config = config.copy()
        kw: dict[str, Any] = dict()
        cls._parse_config(config, kw)
        process_index: int = kw["process_index"]
        actor_pool_config = kw["config"]  # type: ActorPoolConfig
        cur_pool_config = actor_pool_config.get_pool_config(process_index)
        external_addresses = cur_pool_config["external_address"]
        internal_address = kw["internal_address"]

        # import predefined modules
        modules = cur_pool_config["modules"] or []
        for mod in modules:
            __import__(mod, globals(), locals(), [])
        # make sure all lazy imports loaded
        with _disable_log_temporally():
            TypeDispatcher.reload_all_lazy_handlers()

        def handle_channel(channel):
            return pool.on_new_channel(channel)

        # create servers
        server_addresses = list(external_addresses)
        if internal_address:
            server_addresses.append(internal_address)
        server_addresses.append(gen_local_address(process_index))
        server_addresses = sorted(set(server_addresses))
        servers = await cls._create_servers(
            server_addresses, handle_channel, actor_pool_config.get_comm_config()
        )
        cls._update_stored_addresses(servers, server_addresses, actor_pool_config, kw)

        # set default router
        # actor context would be able to use exact client
        cls._set_global_router(kw["router"])

        # create pool
        pool = cls(**kw)
        return pool  # type: ignore


ActorPoolType = TypeVar("ActorPoolType", bound=AbstractActorPool)
MainActorPoolType = TypeVar("MainActorPoolType", bound="MainActorPoolBase")
SubProcessHandle = asyncio.subprocess.Process


class SubActorPoolBase(ActorPoolBase):
    __slots__ = ("_main_address", "_watch_main_pool_task")
    _watch_main_pool_task: Optional[asyncio.Task]

    def __init__(
        self,
        process_index: int,
        label: str,
        external_address: str,
        internal_address: str,
        env: dict,
        router: Router,
        config: ActorPoolConfig,
        servers: list[Server],
        main_address: str,
        main_pool_pid: Optional[int],
    ):
        super().__init__(
            process_index,
            label,
            external_address,
            internal_address,
            env,
            router,
            config,
            servers,
        )
        self._main_address = main_address
        if main_pool_pid:
            self._watch_main_pool_task = asyncio.create_task(
                self._watch_main_pool(main_pool_pid)
            )
        else:
            self._watch_main_pool_task = None

    async def _watch_main_pool(self, main_pool_pid: int):
        main_process = psutil.Process(main_pool_pid)
        while not self.stopped:
            try:
                await asyncio.to_thread(main_process.status)
                await asyncio.sleep(0.1)
                continue
            except (psutil.NoSuchProcess, ProcessLookupError, asyncio.CancelledError):
                # main pool died
                break

        if not self.stopped:
            await self.stop()

    async def notify_main_pool_to_destroy(
        self, message: DestroyActorMessage
    ):  # pragma: no cover
        await self.call(self._main_address, message)

    async def notify_main_pool_to_create(self, message: CreateActorMessage):
        reg_message = ControlMessage(
            new_message_id(),
            self.external_address,
            ControlMessageType.add_sub_pool_actor,
            (self.external_address, message.allocate_strategy, message),
        )
        await self.call(self._main_address, reg_message)

    @implements(AbstractActorPool.create_actor)
    async def create_actor(self, message: CreateActorMessage) -> ResultMessageType:
        result = await super().create_actor(message)
        if not message.from_main:
            await self.notify_main_pool_to_create(message)
        return result

    @implements(AbstractActorPool.actor_ref)
    async def actor_ref(self, message: ActorRefMessage) -> ResultMessageType:
        result = await super().actor_ref(message)
        if isinstance(result, ErrorMessage):
            # need a new message id to call main actor
            main_message = ActorRefMessage(
                new_message_id(),
                create_actor_ref(self._main_address, message.actor_ref.uid),
            )
            result = await self.call(self._main_address, main_message)
            # rewrite to message_id of the original request
            result.message_id = message.message_id
        return result

    @implements(AbstractActorPool.destroy_actor)
    async def destroy_actor(self, message: DestroyActorMessage) -> ResultMessageType:
        result = await super().destroy_actor(message)
        if isinstance(result, ResultMessage) and not message.from_main:
            # sync back to main actor pool
            await self.notify_main_pool_to_destroy(message)
        return result

    @implements(AbstractActorPool.handle_control_command)
    async def handle_control_command(
        self, message: ControlMessage
    ) -> ResultMessageType:
        if message.control_message_type == ControlMessageType.sync_config:
            self._main_address = message.address
        return await super().handle_control_command(message)

    @staticmethod
    def _parse_config(config: dict, kw: dict) -> dict:
        main_pool_pid = config.pop("main_pool_pid", None)
        kw = AbstractActorPool._parse_config(config, kw)
        pool_config: ActorPoolConfig = kw["config"]
        main_process_index = pool_config.get_process_indexes()[0]
        kw["main_address"] = pool_config.get_pool_config(main_process_index)[
            "external_address"
        ][0]
        kw["main_pool_pid"] = main_pool_pid
        return kw

    async def stop(self):
        await super().stop()
        if self._watch_main_pool_task:
            self._watch_main_pool_task.cancel()
            await self._watch_main_pool_task


class MainActorPoolBase(ActorPoolBase):
    __slots__ = (
        "_allocated_actors",
        "sub_actor_pool_manager",
        "_auto_recover",
        "_monitor_task",
        "_on_process_down",
        "_on_process_recover",
        "_recover_events",
        "_allocation_lock",
        "sub_processes",
    )

    def __init__(
        self,
        process_index: int,
        label: str,
        external_address: str,
        internal_address: str,
        env: dict,
        router: Router,
        config: ActorPoolConfig,
        servers: list[Server],
        auto_recover: str | bool = "actor",
        on_process_down: Callable[[MainActorPoolType, str], None] | None = None,
        on_process_recover: Callable[[MainActorPoolType, str], None] | None = None,
    ):
        super().__init__(
            process_index,
            label,
            external_address,
            internal_address,
            env,
            router,
            config,
            servers,
        )

        # auto recovering
        self._auto_recover = auto_recover
        self._monitor_task: Optional[asyncio.Task] = None
        self._on_process_down = on_process_down
        self._on_process_recover = on_process_recover
        self._recover_events: dict[str, asyncio.Event] = dict()

        # states
        self._allocated_actors: allocated_type = {
            addr: dict() for addr in self._config.get_external_addresses()
        }
        self._allocation_lock = threading.Lock()

        self.sub_processes: dict[str, SubProcessHandle] = dict()

    _process_index_gen = itertools.count()

    @classmethod
    def process_index_gen(cls, address):
        # make sure different processes does not share process indexes
        pid = os.getpid()
        for idx in cls._process_index_gen:
            yield pid << 16 + idx

    @property
    def _sub_processes(self):
        return self.sub_processes

    @implements(AbstractActorPool.create_actor)
    async def create_actor(self, message: CreateActorMessage) -> ResultMessageType:
        with _ErrorProcessor(
            address=self.external_address,
            message_id=message.message_id,
            protocol=message.protocol,
        ) as processor:
            allocate_strategy = message.allocate_strategy
            with self._allocation_lock:
                # get allocated address according to corresponding strategy
                address = allocate_strategy.get_allocated_address(
                    self._config, self._allocated_actors
                )
                # set placeholder to make sure this label is occupied
                self._allocated_actors[address][None] = (allocate_strategy, message)
            if address == self.external_address:
                # creating actor on main actor pool
                result = await super().create_actor(message)
                if isinstance(result, ResultMessage):
                    self._allocated_actors[self.external_address][result.result] = (
                        allocate_strategy,
                        message,
                    )
                processor.result = result
            else:
                # creating actor on sub actor pool
                # rewrite allocate strategy to AddressSpecified
                new_allocate_strategy = AddressSpecified(address)
                new_create_actor_message = CreateActorMessage(
                    message.message_id,
                    message.actor_cls,
                    message.actor_id,
                    message.args,
                    message.kwargs,
                    allocate_strategy=new_allocate_strategy,
                    from_main=True,
                    protocol=message.protocol,
                    message_trace=message.message_trace,
                )
                result = await self.call(address, new_create_actor_message)
                if isinstance(result, ResultMessage):
                    self._allocated_actors[address][result.result] = (
                        allocate_strategy,
                        new_create_actor_message,
                    )
                processor.result = result

            # revert placeholder
            self._allocated_actors[address].pop(None, None)

        return processor.result

    @implements(AbstractActorPool.has_actor)
    async def has_actor(self, message: HasActorMessage) -> ResultMessage:
        actor_ref = message.actor_ref
        # lookup allocated
        for address, item in self._allocated_actors.items():
            ref = create_actor_ref(address, to_binary(actor_ref.uid))
            if ref in item:
                return ResultMessage(
                    message.message_id, True, protocol=message.protocol
                )

        return ResultMessage(message.message_id, False, protocol=message.protocol)

    @implements(AbstractActorPool.destroy_actor)
    async def destroy_actor(self, message: DestroyActorMessage) -> ResultMessageType:
        actor_ref_message = ActorRefMessage(
            message.message_id, message.actor_ref, protocol=message.protocol
        )
        result = await self.actor_ref(actor_ref_message)
        if not isinstance(result, ResultMessage):
            return result
        real_actor_ref = result.result
        if real_actor_ref.address == self.external_address:
            result = await super().destroy_actor(message)
            if result.message_type == MessageType.error:
                return result
            del self._allocated_actors[self.external_address][real_actor_ref]
            return ResultMessage(
                message.message_id, real_actor_ref.uid, protocol=message.protocol
            )
        # remove allocated actor ref
        self._allocated_actors[real_actor_ref.address].pop(real_actor_ref, None)
        new_destroy_message = DestroyActorMessage(
            message.message_id,
            real_actor_ref,
            from_main=True,
            protocol=message.protocol,
        )
        return await self.call(real_actor_ref.address, new_destroy_message)

    @implements(AbstractActorPool.send)
    async def send(self, message: SendMessage) -> ResultMessageType:
        if message.actor_ref.uid in self._actors:
            return await super().send(message)
        actor_ref_message = ActorRefMessage(
            message.message_id, message.actor_ref, protocol=message.protocol
        )
        result = await self.actor_ref(actor_ref_message)
        if not isinstance(result, ResultMessage):
            return result
        actor_ref = result.result
        new_send_message = SendMessage(
            message.message_id,
            actor_ref,
            message.content,
            protocol=message.protocol,
            message_trace=message.message_trace,
        )
        return await self.call(actor_ref.address, new_send_message)

    @implements(AbstractActorPool.tell)
    async def tell(self, message: TellMessage) -> ResultMessageType:
        if message.actor_ref.uid in self._actors:
            return await super().tell(message)
        actor_ref_message = ActorRefMessage(
            message.message_id, message.actor_ref, protocol=message.protocol
        )
        result = await self.actor_ref(actor_ref_message)
        if not isinstance(result, ResultMessage):
            return result
        actor_ref = result.result
        new_tell_message = TellMessage(
            message.message_id,
            actor_ref,
            message.content,
            protocol=message.protocol,
            message_trace=message.message_trace,
        )
        return await self.call(actor_ref.address, new_tell_message)

    @implements(AbstractActorPool.actor_ref)
    async def actor_ref(self, message: ActorRefMessage) -> ResultMessageType:
        actor_ref = message.actor_ref
        actor_ref.uid = to_binary(actor_ref.uid)
        if actor_ref.address == self.external_address and actor_ref.uid in self._actors:
            actor_ref.proxy_addresses = self._router.get_proxies(actor_ref.address)
            return ResultMessage(
                message.message_id, actor_ref, protocol=message.protocol
            )

        # lookup allocated
        for address, item in self._allocated_actors.items():
            ref = create_actor_ref(address, actor_ref.uid)
            if ref in item:
                ref.proxy_addresses = self._router.get_proxies(ref.address)
                return ResultMessage(message.message_id, ref, protocol=message.protocol)

        with _ErrorProcessor(
            self.external_address, message.message_id, protocol=message.protocol
        ) as processor:
            raise ActorNotExist(
                f"Actor {actor_ref.uid} does not exist in {actor_ref.address}"
            )

        return processor.result

    @implements(AbstractActorPool.cancel)
    async def cancel(self, message: CancelMessage) -> ResultMessageType:
        if message.address == self.external_address:
            # local message
            return await super().cancel(message)
        # redirect to sub pool
        return await self.call(message.address, message)

    @implements(AbstractActorPool.handle_control_command)
    async def handle_control_command(
        self, message: ControlMessage
    ) -> ResultMessageType:
        with _ErrorProcessor(
            self.external_address, message.message_id, message.protocol
        ) as processor:
            if message.address == self.external_address:
                if message.control_message_type == ControlMessageType.sync_config:
                    # sync config, need to notify all sub pools
                    tasks = []
                    for addr in self.sub_processes:
                        control_message = ControlMessage(
                            new_message_id(),
                            message.address,
                            message.control_message_type,
                            message.content,
                            protocol=message.protocol,
                            message_trace=message.message_trace,
                        )
                        tasks.append(
                            asyncio.create_task(self.call(addr, control_message))
                        )
                    # call super
                    task = asyncio.create_task(super().handle_control_command(message))
                    tasks.append(task)
                    await asyncio.gather(*tasks)
                    processor.result = await task
                else:
                    processor.result = await super().handle_control_command(message)
            elif message.control_message_type == ControlMessageType.stop:
                timeout, force = (
                    message.content if message.content is not None else (None, False)
                )
                await self.stop_sub_pool(
                    message.address,
                    self.sub_processes[message.address],
                    timeout=timeout,
                    force=force,
                )
                processor.result = ResultMessage(
                    message.message_id, True, protocol=message.protocol
                )
            elif message.control_message_type == ControlMessageType.wait_pool_recovered:
                if self._auto_recover and message.address not in self._recover_events:
                    self._recover_events[message.address] = asyncio.Event()

                event = self._recover_events.get(message.address, None)
                if event is not None:
                    await event.wait()
                processor.result = ResultMessage(
                    message.message_id, True, protocol=message.protocol
                )
            elif message.control_message_type == ControlMessageType.add_sub_pool_actor:
                address, allocate_strategy, create_message = message.content
                create_message.from_main = True
                ref = create_actor_ref(address, to_binary(create_message.actor_id))
                self._allocated_actors[address][ref] = (
                    allocate_strategy,
                    create_message,
                )
                processor.result = ResultMessage(
                    message.message_id, True, protocol=message.protocol
                )
            else:
                processor.result = await self.call(message.address, message)
        return processor.result

    @staticmethod
    def _parse_config(config: dict, kw: dict) -> dict:
        kw["auto_recover"] = config.pop("auto_recover", "actor")
        kw["on_process_down"] = config.pop("on_process_down", None)
        kw["on_process_recover"] = config.pop("on_process_recover", None)
        kw = AbstractActorPool._parse_config(config, kw)
        return kw

    @classmethod
    @implements(AbstractActorPool.create)
    async def create(cls, config: dict) -> MainActorPoolType:
        config = config.copy()
        actor_pool_config: ActorPoolConfig = config.get("actor_pool_config")  # type: ignore
        if "process_index" not in config:
            config["process_index"] = actor_pool_config.get_process_indexes()[0]
        curr_process_index = config.get("process_index")
        old_config_addresses = set(actor_pool_config.get_external_addresses())

        tasks = []
        subpool_process_idxes = []
        # create sub actor pools
        n_sub_pool = actor_pool_config.n_pool - 1
        if n_sub_pool > 0:
            process_indexes = actor_pool_config.get_process_indexes()
            for process_index in process_indexes:
                if process_index == curr_process_index:
                    continue
                create_pool_task = asyncio.create_task(
                    cls.start_sub_pool(actor_pool_config, process_index)
                )
                await asyncio.sleep(0)
                # await create_pool_task
                tasks.append(create_pool_task)
                subpool_process_idxes.append(process_index)

        processes, ext_addresses = await cls.wait_sub_pools_ready(tasks)
        if ext_addresses:
            for process_index, ext_address in zip(subpool_process_idxes, ext_addresses):
                actor_pool_config.reset_pool_external_address(
                    process_index, ext_address
                )

        # create main actor pool
        pool: MainActorPoolType = await super().create(config)
        addresses = actor_pool_config.get_external_addresses()[1:]

        assert len(addresses) == len(
            processes
        ), f"addresses {addresses}, processes {processes}"
        for addr, proc in zip(addresses, processes):
            pool.attach_sub_process(addr, proc)

        new_config_addresses = set(actor_pool_config.get_external_addresses())
        if old_config_addresses != new_config_addresses:
            control_message = ControlMessage(
                message_id=new_message_id(),
                address=pool.external_address,
                control_message_type=ControlMessageType.sync_config,
                content=actor_pool_config,
            )
            await pool.handle_control_command(control_message)

        return pool

    async def start_monitor(self):
        # Only start monitor if there are sub processes to monitor
        # This prevents hanging when n_process=0
        if self._monitor_task is None and self.sub_processes:
            self._monitor_task = asyncio.create_task(self.monitor_sub_pools())
        return self._monitor_task

    @implements(AbstractActorPool.stop)
    async def stop(self):
        global_router = Router.get_instance()
        if global_router is not None:
            global_router.remove_router(self._router)

        # turn off auto recover to avoid errors
        self._auto_recover = False
        self._stopped.set()
        if self._monitor_task and not self._monitor_task.done():
            # Cancel the monitor task to ensure it exits immediately
            self._monitor_task.cancel()
            try:
                await self._monitor_task
            except asyncio.CancelledError:
                pass  # Expected when cancelling the task
            self._monitor_task = None
        await self.stop_sub_pools()
        await super().stop()

    @classmethod
    @abstractmethod
    async def start_sub_pool(
        cls,
        actor_pool_config: ActorPoolConfig,
        process_index: int,
        start_python: str | None = None,
    ):
        """Start a sub actor pool"""

    @classmethod
    @abstractmethod
    async def wait_sub_pools_ready(cls, create_pool_tasks: list[asyncio.Task]):
        """Wait all sub pools ready"""

    def attach_sub_process(self, external_address: str, process: SubProcessHandle):
        self.sub_processes[external_address] = process

    async def stop_sub_pools(self):
        to_stop_processes: dict[str, SubProcessHandle] = dict()  # type: ignore
        for address, process in self.sub_processes.items():
            if not await self.is_sub_pool_alive(process):
                continue
            to_stop_processes[address] = process

        tasks = []
        for address, process in to_stop_processes.items():
            tasks.append(self.stop_sub_pool(address, process))
        await asyncio.gather(*tasks)

    async def stop_sub_pool(
        self,
        address: str,
        process: SubProcessHandle,
        timeout: float | None = None,
        force: bool = False,
    ):
        if force:
            await self.kill_sub_pool(process, force=True)
            return

        stop_message = ControlMessage(
            new_message_id(),
            address,
            ControlMessageType.stop,
            None,
            protocol=DEFAULT_PROTOCOL,
        )
        try:
            if timeout is None:
                # Use a short timeout for graceful shutdown to avoid hanging
                timeout = 2.0

            call = asyncio.create_task(self.call(address, stop_message))
            try:
                await asyncio.wait_for(call, timeout)
            except (futures.TimeoutError, asyncio.TimeoutError):
                force = True
        except (ConnectionError, ServerClosed):
            # process dead maybe, ignore it
            force = True
        # kill process
        await self.kill_sub_pool(process, force=force)

    @abstractmethod
    async def kill_sub_pool(self, process: SubProcessHandle, force: bool = False):
        """Kill a sub actor pool"""

    @abstractmethod
    async def is_sub_pool_alive(self, process: SubProcessHandle):
        """
        Check whether sub pool process is alive
        Parameters
        ----------
        process : SubProcessHandle
            sub pool process handle
        Returns
        -------
        bool
        """

    @abstractmethod
    def recover_sub_pool(self, address):
        """Recover a sub actor pool"""

    def process_sub_pool_lost(self, address: str):
        if self._auto_recover in (False, "process"):
            # process down, when not auto_recover
            # or only recover process, remove all created actors
            self._allocated_actors[address] = dict()

    async def monitor_sub_pools(self):
        try:
            while not self._stopped.is_set():
                # Copy sub_processes to avoid changes during recover.
                for address, process in list(self.sub_processes.items()):
                    try:
                        recover_events_discovered = address in self._recover_events
                        if not await self.is_sub_pool_alive(
                            process
                        ):  # pragma: no cover
                            if self._on_process_down is not None:
                                self._on_process_down(self, address)
                            self.process_sub_pool_lost(address)
                            if self._auto_recover:
                                await self.recover_sub_pool(address)
                                if self._on_process_recover is not None:
                                    self._on_process_recover(self, address)
                        if recover_events_discovered:
                            event = self._recover_events.pop(address)
                            event.set()
                    except asyncio.CancelledError:
                        raise
                    except RuntimeError as ex:  # pragma: no cover
                        if "cannot schedule new futures" not in str(ex):
                            # to silence log when process exit, otherwise it
                            # will raise "RuntimeError: cannot schedule new futures
                            # after interpreter shutdown".
                            logger.exception("Monitor sub pool %s failed", address)
                    except Exception:
                        # log the exception instead of stop monitoring the
                        # sub pool silently.
                        logger.exception("Monitor sub pool %s failed", address)

                # check every half second
                await asyncio.sleep(0.5)
        except asyncio.CancelledError:  # pragma: no cover
            # cancelled
            return

    @classmethod
    @abstractmethod
    def get_external_addresses(
        cls,
        address: str,
        n_process: int | None = None,
        ports: list[int] | None = None,
        schemes: list[Optional[str]] | None = None,
    ):
        """Returns external addresses for n pool processes"""

    @classmethod
    @abstractmethod
    def gen_internal_address(
        cls, process_index: int, external_address: str | None = None
    ) -> str | None:
        """Returns internal address for pool of specified process index"""


async def create_actor_pool(
    address: str,
    *,
    pool_cls: Type[MainActorPoolType] | None = None,
    n_process: int | None = None,
    labels: list[str] | None = None,
    ports: list[int] | None = None,
    envs: list[dict] | None = None,
    external_address_schemes: list[Optional[str]] | None = None,
    enable_internal_addresses: list[bool] | None = None,
    auto_recover: str | bool = "actor",
    modules: list[str] | None = None,
    suspend_sigint: bool | None = None,
    use_uvloop: str | bool = "auto",
    logging_conf: dict | None = None,
    proxy_conf: dict | None = None,
    on_process_down: Callable[[MainActorPoolType, str], None] | None = None,
    on_process_recover: Callable[[MainActorPoolType, str], None] | None = None,
    extra_conf: dict | None = None,
    **kwargs,
) -> MainActorPoolType:
    if n_process is None:
        n_process = multiprocessing.cpu_count()
    if labels and len(labels) != n_process + 1:
        raise ValueError(
            f"`labels` should be of size {n_process + 1}, got {len(labels)}"
        )
    if envs and len(envs) != n_process:
        raise ValueError(f"`envs` should be of size {n_process}, got {len(envs)}")
    if external_address_schemes and len(external_address_schemes) != n_process + 1:
        raise ValueError(
            f"`external_address_schemes` should be of size {n_process + 1}, "
            f"got {len(external_address_schemes)}"
        )
    if enable_internal_addresses and len(enable_internal_addresses) != n_process + 1:
        raise ValueError(
            f"`enable_internal_addresses` should be of size {n_process + 1}, "
            f"got {len(enable_internal_addresses)}"
        )
    elif not enable_internal_addresses:
        enable_internal_addresses = [True] * (n_process + 1)
    if auto_recover is True:
        auto_recover = "actor"
    if auto_recover not in ("actor", "process", False):
        raise ValueError(
            f'`auto_recover` should be one of "actor", "process", '
            f"True or False, got {auto_recover}"
        )
    if use_uvloop == "auto":
        try:
            import uvloop  # noqa: F401 # pylint: disable=unused-variable

            use_uvloop = True
        except ImportError:
            use_uvloop = False

    assert pool_cls is not None
    external_addresses = pool_cls.get_external_addresses(
        address, n_process=n_process, ports=ports, schemes=external_address_schemes
    )
    actor_pool_config = ActorPoolConfig()
    actor_pool_config.add_metric_configs(kwargs.get("metrics", {}))
    # add proxy config
    actor_pool_config.add_proxy_config(proxy_conf)
    # add main config
    process_index_gen = pool_cls.process_index_gen(address)
    main_process_index = next(process_index_gen)
    main_internal_address = (
        pool_cls.gen_internal_address(main_process_index, external_addresses[0])
        if enable_internal_addresses[0]
        else None
    )
    actor_pool_config.add_pool_conf(
        main_process_index,
        labels[0] if labels else None,
        main_internal_address,
        external_addresses[0],
        modules=modules,
        suspend_sigint=suspend_sigint,
        use_uvloop=use_uvloop,  # type: ignore
        logging_conf=logging_conf,
        kwargs=kwargs,
    )
    # add sub configs
    for i in range(n_process):
        sub_process_index = next(process_index_gen)
        internal_address = (
            pool_cls.gen_internal_address(sub_process_index, external_addresses[i + 1])
            if enable_internal_addresses[i + 1]
            else None
        )
        actor_pool_config.add_pool_conf(
            sub_process_index,
            labels[i + 1] if labels else None,
            internal_address,
            external_addresses[i + 1],
            env=envs[i] if envs else None,
            modules=modules,
            suspend_sigint=suspend_sigint,
            use_uvloop=use_uvloop,  # type: ignore
            logging_conf=logging_conf,
            kwargs=kwargs,
        )
    actor_pool_config.add_comm_config(extra_conf)

    pool: MainActorPoolType = await pool_cls.create(
        {
            "actor_pool_config": actor_pool_config,
            "process_index": main_process_index,
            "auto_recover": auto_recover,
            "on_process_down": on_process_down,
            "on_process_recover": on_process_recover,
        }
    )
    await pool.start()
    return pool



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from ...backend import register_backend
from ..indigen.backend import IndigenActorBackend
from .pool import TestMainActorPool


@register_backend
class TestActorBackend(IndigenActorBackend):
    @staticmethod
    def name():
        return "test"

    @classmethod
    async def create_actor_pool(
        cls, address: str, n_process: int | None = None, **kwargs
    ):
        from ..pool import create_actor_pool

        assert n_process is not None
        return await create_actor_pool(
            address, pool_cls=TestMainActorPool, n_process=n_process, **kwargs
        )



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# We need to extend cupy's inner class because an actor is a daemonic processes
# which are not allowed to have children. However, the origin code in cupy
# will create children processes.

import queue
import socket
import threading
from ctypes import sizeof

from ...utils import lazy_import

cupy = lazy_import("cupy")

if cupy is not None:
    import cupyx.distributed
    from cupy.cuda import nccl
    from cupyx.distributed import _klv_utils, _store, _store_actions

    class ExceptionAwareThreading(threading.Thread):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._exception = None
            self.q = queue.Queue()

        def run(self):
            try:
                super().run()
                self.q.put(None)
            except Exception as e:
                self.q.put(e)

        def join(self):
            super().join()
            if not self.q.empty():
                exception = self.q.get()
                if exception is not None:
                    raise exception

    class TCPStore:
        # This is only used for initialization of nccl so we don't care
        # too much about performance
        def __init__(self, world_size):
            self.storage = {}
            self._thread = None
            self._world_size = world_size
            self._run = 1
            # For implementing a barrier
            self._lock = threading.Lock()
            self._current_barrier = None

        def __del__(self):
            if not _store._exit_mode:
                self.stop()

        def _thread_request(self, c_socket):
            with c_socket:
                # Receive in KLV format
                action_bytes = c_socket.recv(sizeof(_klv_utils.action_t))
                if len(action_bytes) > 0:
                    action_m = _klv_utils.action_t.from_buffer_copy(action_bytes)
                    if action_m.length > 256:
                        raise ValueError("Invalid length for message")
                    value = bytearray(action_m.value)[: action_m.length]
                    r = _store_actions.execute_action(action_m.action, value, self)
                    if r is not None:
                        c_socket.sendall(r.klv())

        def _server_loop(self, host, port):
            # This is for minimum info exchange during initialization
            # a single connection allows to implement locking mechanics easily
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                s.bind((host, port))
                s.listen()
                s.settimeout(0.5)
                while self._run == 1:
                    try:
                        c_socket, addr = s.accept()
                    except socket.timeout:
                        continue

                    t = threading.Thread(
                        target=self._thread_request, args=(c_socket,), daemon=True
                    )
                    t.start()

        def run(self, host=_store._DEFAULT_HOST, port=_store._DEFAULT_PORT):
            # Run the TCP store in a different process
            t = ExceptionAwareThreading(target=self._server_loop, args=(host, port))
            t.start()
            self._thread = t

        def stop(self):
            if _store._exit_mode:
                return  # Prevent shutdown errors
            if self._thread is not None:
                # acquire the lock
                self._lock.acquire()
                self._run = 0
                self._lock.release()
                self._thread.join()

    class XoscarNCCLBackend(cupyx.distributed.NCCLBackend):
        """Interface that uses NVIDIA's NCCL to perform communications.

        Args:
            n_devices (int): Total number of devices that will be used in the
                distributed execution.
            rank (int): Unique id of the GPU that the communicator is associated to
                its value needs to be `0 <= rank < n_devices`.
            host (str, optional): host address for the process rendezvous on
                initialization. Defaults to `"127.0.0.1"`.
            port (int, optional): port used for the process rendezvous on
                initialization. Defaults to `13333`.
            use_mpi(bool, optional): switch between MPI and use the included TCP
                server for initialization & synchronization. Defaults to `False`.
        """

        def __init__(
            self,
            n_devices,
            rank,
            tcpstore,
            host=_store._DEFAULT_HOST,
            port=_store._DEFAULT_PORT,
            use_mpi=False,
        ):
            self._tcpstore = tcpstore
            super().__init__(n_devices, rank, host, port, use_mpi)

        def _init_with_tcp_store(self, n_devices, rank, host, port):
            nccl_id = None
            if rank == 0:
                self._tcpstore.run(host, port)
                nccl_id = nccl.get_unique_id()
                # get_unique_id return negative values due to cython issues
                # with bytes && c strings. We shift them by 128 to
                # make them positive and send them as bytes to the proxy store
                shifted_nccl_id = bytes([b + 128 for b in nccl_id])
                self._store_proxy["nccl_id"] = shifted_nccl_id
                self._store_proxy.barrier()
            else:
                self._store_proxy.barrier()
                nccl_id = self._store_proxy["nccl_id"]
                nccl_id = tuple([int(b) - 128 for b in nccl_id])
            self._comm = nccl.NcclCommunicator(n_devices, nccl_id, rank)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import asyncio.subprocess
import configparser
import itertools
import logging.config
import os
import pickle
import random
import signal
import struct
import sys
import threading
import time
import uuid
from enum import IntEnum
from typing import List, Optional

import psutil

from ..._utils import reset_id_random_seed
from ...utils import ensure_coverage
from ..config import ActorPoolConfig
from ..message import (
    ControlMessage,
    ControlMessageType,
    CreateActorMessage,
    new_message_id,
)
from ..pool import MainActorPoolBase, SubActorPoolBase, _register_message_handler
from . import shared_memory
from .fate_sharing import create_subprocess_exec

_SUBPROCESS_SHM_SIZE = 10240
_is_windows: bool = sys.platform.startswith("win")

logger = logging.getLogger(__name__)


class _ShmSeq(IntEnum):
    INIT_PARAMS = 1
    INIT_RESULT = 2


def _shm_put_object(seq: _ShmSeq, shm: shared_memory.SharedMemory, o: object):
    serialized = pickle.dumps(o)
    assert (
        len(serialized) < _SUBPROCESS_SHM_SIZE - 8
    ), f"Serialized object {o} is too long."
    shm.buf[4:8] = struct.pack("<I", len(serialized))
    shm.buf[8 : 8 + len(serialized)] = serialized
    shm.buf[:4] = struct.pack("<I", seq)


def _shm_get_object(seq: _ShmSeq, shm: shared_memory.SharedMemory):
    recv_seq = struct.unpack("<I", shm.buf[:4])[0]
    if recv_seq != seq:
        return
    size = struct.unpack("<I", shm.buf[4:8])[0]
    return pickle.loads(shm.buf[8 : 8 + size])


@_register_message_handler
class MainActorPool(MainActorPoolBase):
    @classmethod
    def get_external_addresses(
        cls,
        address: str,
        n_process: int | None = None,
        ports: list[int] | None = None,
        schemes: list[Optional[str]] | None = None,
    ):
        """Get external address for every process"""
        assert n_process is not None
        if ":" in address:
            host, port_str = address.rsplit(":", 1)
            port = int(port_str)
            if ports:
                if len(ports) != n_process:
                    raise ValueError(
                        f"`ports` specified, but its count "
                        f"is not equal to `n_process`, "
                        f"number of ports: {len(ports)}, "
                        f"n_process: {n_process}"
                    )
                sub_ports = ports
            else:
                sub_ports = [0] * n_process
        else:
            host = address
            if ports and len(ports) != n_process + 1:
                # ports specified, the first of which should be main port
                raise ValueError(
                    f"`ports` specified, but its count "
                    f"is not equal to `n_process` + 1, "
                    f"number of ports: {len(ports)}, "
                    f"n_process + 1: {n_process + 1}"
                )
            elif not ports:
                ports = [0] * (n_process + 1)
            port = ports[0]
            sub_ports = ports[1:]
        if not schemes:
            prefix_iter = itertools.repeat("")
        else:
            prefix_iter = [f"{scheme}://" if scheme else "" for scheme in schemes]  # type: ignore
        return [
            f"{prefix}{host}:{port}"
            for port, prefix in zip([port] + sub_ports, prefix_iter)
        ]

    @classmethod
    def gen_internal_address(
        cls, process_index: int, external_address: str | None = None
    ) -> str | None:
        if hasattr(asyncio, "start_unix_server"):
            return f"unixsocket:///{process_index}"
        else:
            return external_address

    @classmethod
    async def start_sub_pool(
        cls,
        actor_pool_config: ActorPoolConfig,
        process_index: int,
        start_python: str | None = None,
    ):
        return await cls._create_sub_pool_from_parent(
            actor_pool_config, process_index, start_python
        )

    @classmethod
    async def wait_sub_pools_ready(cls, create_pool_tasks: List[asyncio.Task]):
        processes: list[asyncio.subprocess.Process] = []
        ext_addresses = []
        error = None
        for task in create_pool_tasks:
            process, address = await task
            processes.append(process)
            ext_addresses.append(address)
        if error:
            for p in processes:
                # error happens, kill all subprocesses
                p.kill()
            raise error
        return processes, ext_addresses

    @classmethod
    def _start_sub_pool_in_child(
        cls,
        shm_name: str,
    ):
        ensure_coverage()

        shm = shared_memory.SharedMemory(shm_name, track=False)
        try:
            config = _shm_get_object(_ShmSeq.INIT_PARAMS, shm)
            # Check Python version once.
            sub_pool_python_version = config.pop("python_version", None)
            if (
                sub_pool_python_version is not None
                and sub_pool_python_version != sys.hexversion
            ):
                logger.warning(
                    f"The sub pool is using a different Python version, you may encounter serialization issues."
                    f" sub pool: {sub_pool_python_version}, main pool: {sys.hexversion}"
                )
            actor_config = config["actor_pool_config"]
            process_index = config["process_index"]
            main_pool_pid = config["main_pool_pid"]

            def _check_ppid():
                while True:
                    try:
                        # We can't simply check if the os.getppid() equals with main_pool_pid,
                        # as the double fork may result in a new process as the parent.
                        psutil.Process(main_pool_pid)
                    except psutil.NoSuchProcess:
                        logger.error("Exit due to main pool %s exit.", main_pool_pid)
                        os._exit(233)  # Special exit code for debugging.
                    except Exception as e:
                        logger.exception("Check ppid failed: %s", e)
                    time.sleep(10)

            t = threading.Thread(target=_check_ppid, daemon=True)
            t.start()

            # make sure enough randomness for every sub pool
            random.seed(uuid.uuid1().bytes)
            reset_id_random_seed()

            conf = actor_config.get_pool_config(process_index)
            suspend_sigint = conf["suspend_sigint"]
            if suspend_sigint:
                signal.signal(signal.SIGINT, lambda *_: None)

            logging_conf = conf["logging_conf"] or {}
            if isinstance(logging_conf, configparser.RawConfigParser):
                logging.config.fileConfig(logging_conf)
            elif logging_conf.get("dict"):
                logging.config.dictConfig(logging_conf["dict"])
            elif logging_conf.get("file"):
                logging.config.fileConfig(logging_conf["file"])
            elif logging_conf.get("level"):
                logging.getLogger("__main__").setLevel(logging_conf["level"])
                logging.getLogger("xoscar").setLevel(logging_conf["level"])
                if logging_conf.get("format"):
                    logging.basicConfig(format=logging_conf["format"])

            use_uvloop = conf["use_uvloop"]
            if use_uvloop:
                import uvloop

                asyncio.set_event_loop(uvloop.new_event_loop())
            else:
                asyncio.set_event_loop(asyncio.new_event_loop())

            coro = cls._create_sub_pool(actor_config, process_index, main_pool_pid, shm)
            asyncio.run(coro)
        finally:
            shm.close()

    @classmethod
    async def _create_sub_pool(
        cls,
        actor_config: ActorPoolConfig,
        process_index: int,
        main_pool_pid: int,
        shm: shared_memory.SharedMemory,
    ):
        cur_pool_config = actor_config.get_pool_config(process_index)
        env = cur_pool_config["env"]
        if env:
            os.environ.update(env)
        pool = await SubActorPool.create(
            {
                "actor_pool_config": actor_config,
                "process_index": process_index,
                "main_pool_pid": main_pool_pid,
            }
        )
        await pool.start()
        _shm_put_object(_ShmSeq.INIT_RESULT, shm, cur_pool_config["external_address"])
        await pool.join()

    @staticmethod
    async def _create_sub_pool_from_parent(
        actor_pool_config: ActorPoolConfig,
        process_index: int,
        start_python: str | None = None,
    ):
        # We check the Python version in _shm_get_object to make it faster,
        # as in most cases the Python versions are the same.
        if start_python is None:
            start_python = sys.executable

        external_addresses: List | None = None
        shm = shared_memory.SharedMemory(
            create=True, size=_SUBPROCESS_SHM_SIZE, track=False
        )
        try:
            _shm_put_object(
                _ShmSeq.INIT_PARAMS,
                shm,
                {
                    "actor_pool_config": actor_pool_config,
                    "process_index": process_index,
                    "main_pool_pid": os.getpid(),
                    "python_version": sys.hexversion,
                },
            )
            cmd = [
                start_python,
                "-m",
                "xoscar.backends.indigen",
                "start_sub_pool",
                "-sn",
                shm.name,
            ]
            # We need to inherit the parent environment to ensure the subprocess works correctly on Windows.
            new_env = dict(os.environ)
            env = actor_pool_config.get_pool_config(process_index).get("env") or {}
            new_env.update(env)
            if os.getenv("XOSCAR_CPU_AFFINITY") == "1":
                import multiprocessing

                total_cores = multiprocessing.cpu_count()
                all_cores_range = f"0-{total_cores - 1}"
                cmd = ["taskset", "-c", all_cores_range] + cmd
            logger.info("Creating sub pool via command: %s", cmd)
            process = await create_subprocess_exec(*cmd, env=new_env)

            def _get_external_addresses():
                try:
                    nonlocal external_addresses
                    while (
                        shm
                        and shm.buf is not None
                        and not (
                            external_addresses := _shm_get_object(
                                _ShmSeq.INIT_RESULT, shm
                            )
                        )
                    ):
                        time.sleep(0.1)
                except asyncio.CancelledError:
                    pass

            _, unfinished = await asyncio.wait(
                [
                    asyncio.create_task(process.wait()),
                    asyncio.create_task(asyncio.to_thread(_get_external_addresses)),
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )
            for t in unfinished:
                t.cancel()
        finally:
            shm.close()
            shm.unlink()
        if external_addresses is None:
            raise OSError(f"Start sub pool failed, returncode: {process.returncode}")
        return process, external_addresses

    async def append_sub_pool(
        self,
        label: str | None = None,
        internal_address: str | None = None,
        external_address: str | None = None,
        env: dict | None = None,
        modules: list[str] | None = None,
        suspend_sigint: bool | None = None,
        use_uvloop: bool | None = None,
        logging_conf: dict | None = None,
        start_python: str | None = None,
        kwargs: dict | None = None,
    ):
        # external_address has port 0, subprocess will bind random port.
        external_address = (
            external_address
            or MainActorPool.get_external_addresses(self.external_address, n_process=1)[
                -1
            ]
        )

        # use last process index's logging_conf and use_uv_loop config if not provide
        actor_pool_config = self._config.as_dict()
        last_process_index = self._config.get_process_indexes()[-1]
        last_logging_conf = actor_pool_config["pools"][last_process_index][
            "logging_conf"
        ]
        last_use_uv_loop = actor_pool_config["pools"][last_process_index]["use_uvloop"]
        _logging_conf = logging_conf or last_logging_conf
        _use_uv_loop = use_uvloop if use_uvloop is not None else last_use_uv_loop

        process_index = next(MainActorPool.process_index_gen(external_address))
        internal_address = internal_address or MainActorPool.gen_internal_address(
            process_index, external_address
        )

        self._config.add_pool_conf(
            process_index,
            label,
            internal_address,
            external_address,
            env,
            modules,
            suspend_sigint,
            _use_uv_loop,
            _logging_conf,
            kwargs,
        )

        process, external_addresses = await self._create_sub_pool_from_parent(
            self._config, process_index, start_python
        )

        self._config.reset_pool_external_address(process_index, external_addresses[0])
        self.attach_sub_process(external_addresses[0], process)

        control_message = ControlMessage(
            message_id=new_message_id(),
            address=self.external_address,
            control_message_type=ControlMessageType.sync_config,
            content=self._config,
        )
        await self.handle_control_command(control_message)
        # The actual port will return in process_status.
        return external_addresses[0]

    async def remove_sub_pool(
        self, external_address: str, timeout: float | None = None, force: bool = False
    ):
        process = self.sub_processes[external_address]
        process_index = self._config.get_process_index(external_address)
        del self.sub_processes[external_address]
        self._config.remove_pool_config(process_index)
        await self.stop_sub_pool(external_address, process, timeout, force)

        control_message = ControlMessage(
            message_id=new_message_id(),
            address=self.external_address,
            control_message_type=ControlMessageType.sync_config,
            content=self._config,
        )
        await self.handle_control_command(control_message)

    async def kill_sub_pool(
        self, process: asyncio.subprocess.Process, force: bool = False
    ):
        # First, try to terminate the process gracefully
        if not force:
            try:
                process.terminate()
                # Wait for graceful termination
                try:
                    await asyncio.wait_for(process.wait(), timeout=2.0)
                except asyncio.TimeoutError:
                    # Process didn't terminate gracefully, force kill
                    force = True
            except ProcessLookupError:
                # Process already terminated
                pass

        # Force kill if needed or if graceful termination failed
        if force:
            try:
                process.kill()
            except ProcessLookupError:
                # Process already dead
                pass

        # Ensure process is completely terminated and cleaned up
        try:
            # Wait for process to complete
            if process.returncode is None:
                try:
                    await asyncio.wait_for(process.wait(), timeout=5.0)
                except asyncio.TimeoutError:
                    pass
        except ProcessLookupError:
            # Process already terminated
            pass

        # Python 3.13 specific cleanup for waitpid threads
        if sys.version_info >= (3, 13):
            try:
                # Close the transport to clean up waitpid thread
                if hasattr(process, "_transport") and process._transport:
                    process._transport.close()
                # Also try to close the pipe transport if it exists
                if hasattr(process, "_pipes") and process._pipes:
                    for pipe in process._pipes.values():
                        if hasattr(pipe, "close"):
                            pipe.close()
            except Exception:
                # Ignore errors during cleanup
                pass

        # Additional cleanup using psutil to ensure process tree is terminated
        try:
            p = psutil.Process(process.pid)
            if p.is_running():
                # Kill the entire process tree
                for child in p.children(recursive=True):
                    try:
                        child.kill()
                    except psutil.NoSuchProcess:
                        pass
                p.kill()
                p.wait(timeout=2.0)
        except (psutil.NoSuchProcess, psutil.TimeoutExpired):
            # Process already dead or couldn't be killed
            pass

    async def is_sub_pool_alive(self, process: asyncio.subprocess.Process):
        return process.returncode is None

    async def recover_sub_pool(self, address: str):
        process_index = self._config.get_process_index(address)
        # process dead, restart it
        # remember always use spawn to recover sub pool
        task = asyncio.create_task(self.start_sub_pool(self._config, process_index))
        self.sub_processes[address] = (await self.wait_sub_pools_ready([task]))[0][0]

        if self._auto_recover == "actor":
            # need to recover all created actors
            for _, message in self._allocated_actors[address].values():
                create_actor_message: CreateActorMessage = message  # type: ignore
                await self.call(address, create_actor_message)

    async def start(self):
        await super().start()
        await self.start_monitor()


@_register_message_handler
class SubActorPool(SubActorPoolBase):
    pass



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from ...backend import BaseActorBackend, register_backend
from ..context import IndigenActorContext
from .driver import IndigenActorDriver
from .pool import MainActorPool

__all__ = ["IndigenActorBackend"]


@register_backend
class IndigenActorBackend(BaseActorBackend):
    @staticmethod
    def name():
        # None means Indigen is default scheme
        # ucx can be recognized as Indigen backend as well
        return [None, "ucx"]

    @staticmethod
    def get_context_cls():
        return IndigenActorContext

    @staticmethod
    def get_driver_cls():
        return IndigenActorDriver

    @classmethod
    async def create_actor_pool(
        cls, address: str, n_process: int | None = None, **kwargs
    ):
        from ..pool import create_actor_pool

        assert n_process is not None
        return await create_actor_pool(
            address, pool_cls=MainActorPool, n_process=n_process, **kwargs
        )



# Copyright 2022-2025 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional


def check_cuda_available() -> bool:
    try:
        import torch

        return torch.cuda.is_available()
    except (ImportError, AttributeError):
        return False


def get_cuda_version() -> Optional[str]:
    try:
        import torch

        return torch.version.cuda  # e.g. '12.1'
    except (ImportError, AttributeError):
        return None


def get_cuda_arch() -> Optional[str]:
    try:
        import torch

        major, minor = torch.cuda.get_device_capability()
        return f"sm_{major}{minor}"  # e.g. 'sm_80'
    except (ImportError, AttributeError, AssertionError, RuntimeError):
        # If no cuda available,
        # AssertionError("Torch not compiled with CUDA enabled")
        # will be raised
        return None


def check_npu_available() -> bool:
    try:
        import torch
        import torch_npu  # noqa: F401

        return torch.npu.is_available()
    except ImportError:
        return False



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from concurrent.futures import Executor
from typing import Any, Callable

def buffered(func: Callable) -> Callable: ...
def fast_id(obj: Any) -> int: ...

class Serializer:
    serializer_id: int
    def serial(self, obj: Any, context: dict): ...
    def deserial(self, serialized: tuple, context: dict, subs: list[Any]): ...
    def on_deserial_error(
        self,
        serialized: tuple,
        context: dict,
        subs_serialized: list,
        error_index: int,
        exc: BaseException,
    ): ...
    @classmethod
    def register(cls, obj_type, name: str | None = None): ...
    @classmethod
    def unregister(cls, obj_type): ...

class Placeholder:
    id: int
    callbacks: list[Callable]
    def __init__(self, id_: int): ...
    def __hash__(self): ...
    def __eq__(self, other): ...

def serialize(obj: Any, context: dict | None = None): ...
async def serialize_with_spawn(
    obj: Any,
    context: dict | None = None,
    spawn_threshold: int = 100,
    executor: Executor | None = None,
): ...
def deserialize(headers: list, buffers: list, context: dict | None = None): ...
def pickle_buffers(obj: Any) -> list: ...
def unpickle_buffers(buffers: list) -> Any: ...



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import socket
from typing import Dict, Optional

from ....utils import lazy_import
from ..metric import (
    AbstractCounter,
    AbstractGauge,
    AbstractHistogram,
    AbstractMeter,
    AbstractMetric,
)

pc = lazy_import("prometheus_client", rename="pc")


class PrometheusMetricMixin(AbstractMetric):
    def _init(self):
        # Prometheus metric name must match the regex `[a-zA-Z_:][a-zA-Z0-9_:]*`
        # `.` is a common character in metrics, so here replace it with `:`
        self._name = self._name.replace(".", ":")
        self._tag_keys = self._tag_keys + (
            "host",
            "pid",
        )
        self._tags = {"host": socket.gethostname(), "pid": os.getpid()}
        try:
            self._metric = (
                pc.Gauge(self._name, self._description, self._tag_keys) if pc else None
            )
        except ValueError:  # pragma: no cover
            self._metric = None

    def _record(self, value=1, tags: Optional[Dict[str, str]] = None):
        if self._metric:
            if tags is not None:
                tags.update(self._tags)
            else:
                tags = self._tags
            self._metric.labels(**tags).set(value)


class Counter(PrometheusMetricMixin, AbstractCounter):
    pass


class Gauge(PrometheusMetricMixin, AbstractGauge):
    pass


class Meter(PrometheusMetricMixin, AbstractMeter):
    pass


class Histogram(PrometheusMetricMixin, AbstractHistogram):
    pass



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from abc import ABC, abstractmethod
from typing import Any, List, Optional

from ..utils import is_linux, lazy_import
from . import xoscar_pygloo as xp
from .common import (
    RENDEZVOUS_MASTER_IP_ENV_KEY,
    RENDEZVOUS_MASTER_PORT_ENV_KEY,
    AllReduceAlgorithm,
    AllReduceAlgorithmMappingGloo,
    CollectiveReduceOp,
    ReduceOpMappingGloo,
    TypeMappingGloo,
)
from .utils import convert_data_to_cp_array, convert_data_to_np_array

cupy = lazy_import("cupy")
if cupy is not None:
    from .backend.nccl_backend import TCPStore, XoscarNCCLBackend
    from .common import ReduceOpMappingNCCL, ReduceOpMappingNCCLStr, TypeMappingNCCL


class _World:
    def __init__(self):
        self._store = None
        self._device = None
        self._backend = None
        self._nccl_size = 0

    @property
    def store(self):
        return self._store

    @property
    def device(self):
        return self._device

    @store.setter  # type: ignore
    def store(self, store):
        self._store = store

    @device.setter  # type: ignore
    def device(self, device):
        self._device = device


_world = _World()


class ProcessGroup(ABC):
    class Options:
        master_ip: Optional[str] = None
        master_port: Optional[int] = None

    def __init__(
        self,
        rank: int,
        world_size: int,
        group_name: Optional[str] = None,
        pg_options: Optional[Options] = None,
    ):
        self._rank = rank
        self._world_size = world_size
        self._group_name = group_name
        self._option = pg_options

    @property
    def rank(self):
        return self._rank

    @property
    def world_size(self):
        return self._world_size

    @property
    def name(self):
        return self._group_name

    @property
    def options(self):
        return self._option

    @abstractmethod
    def allreduce(self, *args, **kwargs):
        """All reduce function"""

    @abstractmethod
    def reduce(self, *args, **kwargs):
        """Reduce function"""

    @abstractmethod
    def allgather(self, *args, **kwargs):
        """All gather function"""

    @abstractmethod
    def gather(self, *args, **kwargs):
        """Gather function"""

    @abstractmethod
    def scatter(self, *args, **kwargs):
        """Scatter function"""

    @abstractmethod
    def reduce_scatter(self, *args, **kwargs):
        """Reduce scatter function"""

    @abstractmethod
    def alltoall(self, *args, **kwargs):
        """All to all function"""

    @abstractmethod
    def broadcast(self, *args, **kwargs):
        """Broadcast function"""


class ProcessGroupGloo(ProcessGroup):
    def __init__(
        self,
        ip: str,
        rank: int,
        world_size: int,
        group_name: Optional[str] = None,
        pg_options: Optional[ProcessGroup.Options] = None,
    ):
        super().__init__(rank, world_size, group_name, pg_options)
        if _world.store is None:
            master_ip = (
                pg_options.master_ip
                if pg_options is not None
                else os.environ.get(RENDEZVOUS_MASTER_IP_ENV_KEY, None)
            )
            master_port = (
                pg_options.master_port
                if pg_options is not None
                else os.environ.get(RENDEZVOUS_MASTER_PORT_ENV_KEY, None)
            )
            if master_ip is None or master_port is None:
                raise ValueError("Cannot find master ip or port for rendezvous")

            opt = xp.rendezvous.TCPStoreOptions()
            opt.port = int(master_port)
            opt.numWorkers = world_size
            opt.isServer = rank == 0

            store = xp.rendezvous.TCPStore(master_ip, opt)
            if not is_linux():
                attr = xp.transport.uv.attr(ip)  # type: ignore
                dev = xp.transport.uv.CreateDevice(attr)  # type: ignore
            else:
                attr = xp.transport.tcp.attr(ip)
                dev = xp.transport.tcp.CreateDevice(attr)  # type: ignore
            _world.store = store  # type: ignore
            _world.device = dev  # type: ignore
        else:
            store = _world.store
            dev = _world.device

        prefix_store = xp.rendezvous.PrefixStore(group_name or str(world_size), store)  # type: ignore
        context = xp.rendezvous.Context(rank, world_size)
        context.connectFullMesh(prefix_store, dev)
        self._context = context

    def reduce(
        self,
        send_data: Any,
        recv_data: Any,
        op: CollectiveReduceOp = CollectiveReduceOp.SUM,
        root: Optional[int] = 0,
        tag: Optional[int] = 0,
    ):
        send_buf = convert_data_to_np_array(send_data)
        recv_buf = convert_data_to_np_array(recv_data)
        size = send_buf.size
        dtype = send_buf.dtype
        sendptr = send_buf.ctypes.data
        recvptr = recv_buf.ctypes.data
        gloo_type = TypeMappingGloo[dtype.type]
        xp.reduce(
            self._context,
            sendptr,
            recvptr,
            size,
            gloo_type,
            ReduceOpMappingGloo[op],
            root,
            tag,
        )

    def allreduce(
        self,
        send_data: Any,
        recv_data: Any,
        op: CollectiveReduceOp = CollectiveReduceOp.SUM,
        algorithm: AllReduceAlgorithm = AllReduceAlgorithm.RING,
        tag: Optional[int] = 0,
    ):
        send_buf = convert_data_to_np_array(send_data)
        recv_buf = convert_data_to_np_array(recv_data)
        size = send_buf.size
        dtype = send_buf.dtype
        sendptr = send_buf.ctypes.data
        recvptr = recv_buf.ctypes.data
        gloo_type = TypeMappingGloo[dtype.type]
        xp.allreduce(
            self._context,
            sendptr,
            recvptr,
            size,
            gloo_type,
            ReduceOpMappingGloo[op],
            AllReduceAlgorithmMappingGloo[algorithm],
            tag,  # type: ignore
        )

    def gather(
        self,
        send_data: Any,
        recv_data: Any,
        root: Optional[int] = 0,
        tag: Optional[int] = 0,
    ):
        send_buf = convert_data_to_np_array(send_data)
        recv_buf = convert_data_to_np_array(recv_data)
        size = send_buf.size
        dtype = send_buf.dtype
        sendptr = send_buf.ctypes.data
        recvptr = recv_buf.ctypes.data
        gloo_type = TypeMappingGloo[dtype.type]
        xp.gather(self._context, sendptr, recvptr, size, gloo_type, root, tag)

    def allgather(self, send_data: Any, recv_data: Any, tag: Optional[int] = 0):
        send_buf = convert_data_to_np_array(send_data)
        recv_buf = convert_data_to_np_array(recv_data)
        size = send_buf.size
        dtype = send_buf.dtype
        sendptr = send_buf.ctypes.data
        recvptr = recv_buf.ctypes.data
        gloo_type = TypeMappingGloo[dtype.type]
        xp.allgather(self._context, sendptr, recvptr, size, gloo_type, tag)

    def scatter(
        self,
        send_data: List[Any],
        recv_data: Any,
        root: Optional[int] = 0,
        tag: Optional[int] = 0,
    ):
        send_bufs = [convert_data_to_np_array(d) for d in send_data]
        recv_buf = convert_data_to_np_array(recv_data)
        size = sum([d.size for d in send_bufs])
        dtype = recv_buf.dtype
        sendptrs = [d.ctypes.data for d in send_bufs]
        recvptr = recv_buf.ctypes.data
        gloo_type = TypeMappingGloo[dtype.type]
        xp.scatter(self._context, sendptrs, recvptr, size, gloo_type, root, tag)  # type: ignore

    def reduce_scatter(
        self,
        send_data: Any,
        recv_data: Any,
        recv_elems: List[int],
        op: CollectiveReduceOp = CollectiveReduceOp.SUM,
    ):  # pragma: no cover
        send_buf = convert_data_to_np_array(send_data)
        recv_buf = convert_data_to_np_array(recv_data)
        sendptr = send_buf.ctypes.data
        recvptr = recv_buf.ctypes.data
        size = send_buf.size
        dtype = send_buf.dtype
        gloo_type = TypeMappingGloo[dtype.type]
        xp.reduce_scatter(
            self._context,
            sendptr,
            recvptr,
            size,
            recv_elems,
            gloo_type,
            ReduceOpMappingGloo[op],
        )

    def alltoall(self, send_data: Any, recv_data: Any, tag: Optional[int] = 0):
        send_buf = convert_data_to_np_array(send_data)
        recv_buf = convert_data_to_np_array(recv_data)
        size = send_buf.size
        dtype = send_buf.dtype
        sendptr = send_buf.ctypes.data
        recvptr = recv_buf.ctypes.data
        gloo_type = TypeMappingGloo[dtype.type]
        xp.all_to_all(self._context, sendptr, recvptr, size, gloo_type, tag)

    def broadcast(
        self,
        send_data: Any,
        recv_data: Any,
        root: Optional[int] = 0,
        tag: Optional[int] = 0,
    ):
        if send_data is not None:
            send_buf = convert_data_to_np_array(send_data)
            sendptr = send_buf.ctypes.data
        else:
            sendptr = None
        recv_buf = convert_data_to_np_array(recv_data)
        size = recv_buf.size
        dtype = recv_buf.dtype
        recvptr = recv_buf.ctypes.data
        gloo_type = TypeMappingGloo[dtype.type]
        xp.broadcast(
            self._context,
            recvptr if sendptr is None else sendptr,
            recvptr,
            size,
            gloo_type,
            root,
            tag,
        )


class ProcessGroupNCCL(ProcessGroup):
    def __init__(
        self,
        ip: str,
        rank: int,
        device_id: int,
        world_size: int,
        group_name: Optional[str] = None,
        pg_options: Optional[ProcessGroup.Options] = None,
    ):
        assert (
            cupy != None
        ), "cupy is required when creating a group using nccl as backend."
        from cupy.cuda import nccl

        super().__init__(rank, world_size, group_name, pg_options)
        cupy.cuda.Device(device_id).use()
        if _world._backend is None:
            master_ip = (
                pg_options.master_ip
                if pg_options is not None
                else os.environ.get(RENDEZVOUS_MASTER_IP_ENV_KEY, None)
            )
            master_port = (
                pg_options.master_port
                if pg_options is not None
                else os.environ.get(RENDEZVOUS_MASTER_PORT_ENV_KEY, None)
            )
            if master_ip is None or master_port is None:
                raise ValueError("Cannot find master ip or port for rendezvous")
            store = TCPStore(world_size)
            backend = XoscarNCCLBackend(
                world_size, rank, store, master_ip, int(master_port)
            )
            _world._backend = backend
            _world._nccl_size = world_size
            self._is_world = True
            self._backend = backend
        else:
            self._is_world = False
            if rank == 0:
                commId = nccl.get_unique_id()
                ccid = cupy.array(commId)
                for i in range(1, world_size):
                    _world._backend.send(ccid, i, None)
            else:
                commId = (int(i) for i in range(128))
                commId = tuple(commId)
                ccid = cupy.array(commId, dtype="int64")
                _world._backend.recv(ccid, 0, None)
                commId = tuple(ccid.tolist())
            self._backend = nccl.NcclCommunicator(world_size, commId, rank)

    def reduce(
        self,
        send_buf: Any,
        recv_buf: Any,
        op: CollectiveReduceOp = CollectiveReduceOp.SUM,
        root: Optional[int] = 0,
        stream: Optional[Any] = None,
    ):
        send_buf = convert_data_to_cp_array(send_buf)
        recv_buf = convert_data_to_cp_array(recv_buf)
        dtype = send_buf.dtype
        stream = (
            stream
            if stream is not None and isinstance(stream, cupy.cuda.Stream)
            else cupy.cuda.Stream.null
        )
        if self._is_world:
            self._backend.reduce(
                send_buf, recv_buf, root, ReduceOpMappingNCCLStr[op], stream
            )
        else:
            self._backend.reduce(
                send_buf.data.ptr,
                recv_buf.data.ptr,
                send_buf.size,
                TypeMappingNCCL[dtype.type],
                ReduceOpMappingNCCL[op],
                root,
                stream.ptr,
            )

    def allreduce(
        self,
        send_buf: Any,
        recv_buf: Any,
        op: CollectiveReduceOp = CollectiveReduceOp.SUM,
        stream: Optional[Any] = None,
    ):
        send_buf = convert_data_to_cp_array(send_buf)
        recv_buf = convert_data_to_cp_array(recv_buf)
        dtype = send_buf.dtype
        stream = (
            stream
            if stream is not None and isinstance(stream, cupy.cuda.Stream)
            else cupy.cuda.Stream.null
        )
        if self._is_world:
            self._backend.all_reduce(
                send_buf, recv_buf, ReduceOpMappingNCCLStr[op], stream
            )
        else:
            self._backend.allReduce(
                send_buf.data.ptr,
                recv_buf.data.ptr,
                send_buf.size,
                TypeMappingNCCL[dtype.type],
                ReduceOpMappingNCCL[op],
                stream.ptr,
            )

    def gather(
        self,
        send_buf: Any,
        recv_buf: Any,
        root: Optional[int] = 0,
        stream: Optional[Any] = None,
    ):
        assert (
            send_buf.size * self.world_size == recv_buf.size
        ), "Send_size * world_number must be equal to recv_size"
        send_buf = convert_data_to_cp_array(send_buf)
        recv_buf = convert_data_to_cp_array(recv_buf)
        dtype = send_buf.dtype
        stream = (
            stream
            if stream is not None and isinstance(stream, cupy.cuda.Stream)
            else cupy.cuda.Stream.null
        )
        if self._is_world:
            self._backend.gather(send_buf, recv_buf, root, stream)
        else:
            if self.rank == root:
                cupy.cuda.nccl.groupStart()
                for peer in range(self.world_size):
                    if peer == self.rank:
                        recv_buf[peer : peer + 1] = send_buf.reshape(1, -1)
                    else:
                        self._backend.recv(
                            recv_buf[peer : peer + 1].data.ptr,
                            recv_buf[peer : peer + 1].size,
                            TypeMappingNCCL[dtype.type],
                            peer,
                            stream.ptr,
                        )
                cupy.cuda.nccl.groupEnd()
            else:
                send_buf = send_buf.reshape(1, -1)
                self._backend.send(
                    send_buf.data.ptr,
                    send_buf.size,
                    TypeMappingNCCL[dtype.type],
                    root,
                    stream.ptr,
                )

    def allgather(
        self,
        send_buf: Any,
        recv_buf: Any,
        stream: Optional[Any] = None,
    ):
        send_buf = convert_data_to_cp_array(send_buf)
        recv_buf = convert_data_to_cp_array(recv_buf)
        stream = (
            stream
            if stream is not None and isinstance(stream, cupy.cuda.Stream)
            else cupy.cuda.Stream.null
        )
        dtype = send_buf.dtype
        if self._is_world:
            self._backend.all_gather(send_buf, recv_buf, send_buf.size, stream)
        else:
            self._backend.allGather(
                send_buf.data.ptr,
                recv_buf.data.ptr,
                send_buf.size,
                TypeMappingNCCL[dtype.type],
                stream.ptr,
            )

    def scatter(
        self,
        send_buf: List[Any],
        recv_buf: Any,
        root: Optional[int] = 0,
        stream: Optional[Any] = None,
    ):
        send_buf = [convert_data_to_cp_array(d) for d in send_buf]
        recv_buf = convert_data_to_cp_array(recv_buf)
        stream = (
            stream
            if stream is not None and isinstance(stream, cupy.cuda.Stream)
            else cupy.cuda.Stream.null
        )
        if self._is_world:
            send_buf = cupy.concatenate(send_buf).reshape(self.world_size, -1)
            self._backend.scatter(send_buf, recv_buf, root, stream)
        else:
            if self.rank == root:
                assert (
                    len(send_buf) == self.world_size
                ), "Scatter size must be equal to the size of group"
                cupy.cuda.nccl.groupStart()
                for peer in range(self.world_size):
                    send_data = send_buf[peer]
                    if peer == root:
                        recv_buf[:] = send_data
                    else:
                        dtype = send_data.dtype
                        self._backend.send(
                            send_data.data.ptr,
                            send_data.size,
                            TypeMappingNCCL[dtype.type],
                            peer,
                            stream.ptr,
                        )
                cupy.cuda.nccl.groupEnd()
            else:
                dtype = recv_buf.dtype
                self._backend.recv(
                    recv_buf.data.ptr,
                    recv_buf.size,
                    TypeMappingNCCL[dtype.type],
                    root,
                    stream.ptr,
                )

    def reduce_scatter(
        self,
        send_buf: Any,
        recv_buf: Any,
        recv_elems: List[int],
        op: CollectiveReduceOp = CollectiveReduceOp.SUM,
        stream: Optional[Any] = None,
    ):
        send_buf = convert_data_to_cp_array(send_buf)
        recv_buf = convert_data_to_cp_array(recv_buf)
        dtype = send_buf.dtype
        stream = (
            stream
            if stream is not None and isinstance(stream, cupy.cuda.Stream)
            else cupy.cuda.Stream.null
        )
        if self._is_world:
            self._backend.reduce_scatter(
                send_buf,
                recv_buf,
                recv_elems[self.rank],
                ReduceOpMappingNCCLStr[op],
                stream,
            )
        else:
            self._backend.reduceScatter(
                send_buf.data.ptr,
                recv_buf.data.ptr,
                recv_elems[self.rank],
                TypeMappingNCCL[dtype.type],
                ReduceOpMappingNCCL[op],
                stream.ptr,
            )

    def alltoall(
        self,
        send_buf: Any,
        recv_buf: Any,
        stream: Optional[Any] = None,
    ):
        assert (
            self.world_size == send_buf.shape[0]
        ), "The first dim of send data must be equal to world size."
        assert (
            recv_buf.size == send_buf.size
        ), "The size of send data must be equal to the size of recv data."
        send_buf = convert_data_to_cp_array(send_buf)
        recv_buf = convert_data_to_cp_array(recv_buf)
        dtype = send_buf.dtype
        stream = (
            stream
            if stream is not None and isinstance(stream, cupy.cuda.Stream)
            else cupy.cuda.Stream.null
        )
        if self._is_world:
            self._backend.all_to_all(send_buf, recv_buf, stream)
        else:
            cupy.cuda.nccl.groupStart()
            for peer in range(self.world_size):
                if peer == self.rank:
                    recv_buf[peer : peer + 1] = send_buf[peer : peer + 1]
                else:
                    if self.rank > peer:
                        self._backend.recv(
                            recv_buf[peer : peer + 1].data.ptr,
                            recv_buf[peer : peer + 1].size,
                            TypeMappingNCCL[dtype.type],
                            peer,
                            stream.ptr,
                        )
                        self._backend.send(
                            send_buf[peer : peer + 1].data.ptr,
                            send_buf[peer : peer + 1].size,
                            TypeMappingNCCL[dtype.type],
                            peer,
                            stream.ptr,
                        )
                    else:
                        self._backend.send(
                            send_buf[peer : peer + 1].data.ptr,
                            send_buf[peer : peer + 1].size,
                            TypeMappingNCCL[dtype.type],
                            peer,
                            stream.ptr,
                        )
                        self._backend.recv(
                            recv_buf[peer : peer + 1].data.ptr,
                            recv_buf[peer : peer + 1].size,
                            TypeMappingNCCL[dtype.type],
                            peer,
                            stream.ptr,
                        )
            cupy.cuda.nccl.groupEnd()

    def broadcast(
        self,
        send_buf: Any,
        recv_buf: Any,
        root: Optional[int] = 0,
        stream: Optional[Any] = None,
    ):
        send_buf = convert_data_to_cp_array(send_buf)
        recv_buf = convert_data_to_cp_array(recv_buf)
        dtype = send_buf.dtype
        stream = (
            stream
            if stream is not None and isinstance(stream, cupy.cuda.Stream)
            else cupy.cuda.Stream.null
        )
        if self._is_world:
            if self._rank == root:
                self._backend.broadcast(send_buf, root, stream)
                if recv_buf is not None and (recv_buf != send_buf).any():
                    recv_buf[:] = send_buf
            else:
                self._backend.broadcast(recv_buf, root, stream)
        else:
            self._backend.broadcast(
                send_buf.data.ptr,
                recv_buf.data.ptr,
                send_buf.size,
                TypeMappingNCCL[dtype.type],
                root,
                stream.ptr,
            )



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import Channel, ChannelType, Client, Server
from .core import gen_local_address, get_client_type, get_server_type
from .dummy import DummyChannel, DummyClient, DummyServer
from .socket import (
    SocketChannel,
    SocketClient,
    SocketServer,
    UnixSocketClient,
    UnixSocketServer,
)
from .ucx import (  # noqa: F401 # pylint: disable=unused-import
    UCXChannel,
    UCXClient,
    UCXServer,
)
